'''Posts a comment to a Bitbucket pull request, with optional arguments for 'Jenkins build 
annotations', indicating build status, encasing content in 'diff markdown', and more!
'''
import configargparse
import mdfrag
import requests

import ntsbuildtools.io
from ntsbuildtools.markdown.comments import MarkdownComment


def main(args):
    # Get the core comment message from args -- these options are mutually exclusive (see config_parser)
    comment = MarkdownComment()
    if args.comment_message:
        comment.add_message(args.comment_message)
    elif args.comment_file:
        message_data = ntsbuildtools.io.readfile(args.comment_file)
        comment.add_message(message_data)
    if args.build_annotation:
        comment.add_build_annotation(args.build_id, args.build_url, args.playbook_limit, args.build_status)
    if args.trim:
        comment.trim_comment(args.trim)

    if len(str(comment).strip()) == 0:
        raise ValueError("Some comment contents must be provided -- the comment to be sent was an empty string.")

    # Actually make the request
    try:
        # If the comment will exceed the max comment size, do fragmentation instead!
        if args.max_comment_size and len(str(comment)) > args.max_comment_size:
            comment_fragments = mdfrag.split_markdown(str(comment), args.max_comment_size)
            # Comments are in reverse chronological, so we should render the fragments in reverse.
            comment_fragments.reverse()
            # cp_i => "Comment Part Index"
            cp_i = len(comment_fragments)
            for comment_part in comment_fragments:
                pagination_str = f'> Comment __{cp_i}__ of __{len(comment_fragments)}__.'
                msg = f'{pagination_str}\n{comment_part}'
                cp_i -= 1
                response = requests.post(
                    f'{args.bitbucket_url}/rest/api/1.0/projects/{args.project}/repos/{args.repo}/pull-requests/{args.pull_request_id}/comments',
                    json={'text': msg},
                    auth=(args.user, args.password)
                )
                response.raise_for_status()
        else:
            response = requests.post(
                f'{args.bitbucket_url}/rest/api/1.0/projects/{args.project}/repos/{args.repo}/pull-requests/{args.pull_request_id}/comments',
                json={'text': str(comment) if comment else '*Empty string provided.*'},
                auth=(args.user, args.password)
            )
            response.raise_for_status()
    except requests.exceptions.HTTPError as error:
        print(f"[ERROR] HTTP Error ({error.response.status_code}): {error.response.text}")
        exit(-1)

def config_parser(parser: configargparse.ArgParser):
    # Parse the 'comment type' with mutex logic -- don't allow user to provide more than one type of message from the provided options.
    possible_inputs = parser.add_mutually_exclusive_group(required=True)
    possible_inputs.add('--message', '-m', dest='comment_message',
                        help='Provide the comment as a message on the command line.')
    possible_inputs.add('--file', '-f', dest='comment_file',
                        help='Provide the comment in a file.')
    # Core parsing bits.
    parser.add('--bitbucket-url', env_var='BITBUCKET_URL', help="URL for Bitbucket.")
    parser.add('--user', required=True, env_var='BITBUCKET_USER',
               help="Bitbucket user that will be used to authenticate to Bitbucket.")
    parser.add('--password', required=True, env_var='BITBUCKET_PASSWORD',
               help="Bitbucket password (or Personal Access Token) for the Bitbucket user.")
    parser.add('--project', required=True, env_var='BITBUCKET_PROJECT',
               help='The Bitbucket project key for the project where the pull request exists.')
    parser.add('--repo', required=True, env_var='BITBUCKET_REPO',
               help='The Bitbucket repository slug for the repository where the pull request exists.')
    parser.add('--pull-request-id', required=True, env_var='PR_ID',
               help='The ID of the Bitbucket pull request to be commented on.')
    # Group of arguments to do with formatting.
    formatting = parser.add_argument_group('formatting')
    formatting.add(
        '--max-comment-size', 
        type=int,
        default=32767,  # Default max comment size in Bitbucket is 32768 bytes
        help='Fragment the comment into chunks smaller than this, in bytes.')
    formatting.add(
        '--trim', '--tail',
        type=int,
        help='Only print the last `TRIM` lines of the provided message/file.'
    )
    # Build Annotation (aka 'ba') parser.
    ba_parser = parser.add_argument_group('Build Annotation')
    ba_parser.add('--build-annotation', action='store_true',
                  help='Provide a "build annotation" in the comment, with information about a Jenkins build.')
    ba_parser.add('--playbook-limit', env_var='PLAYBOOK_LIMIT',
                  help='A string indicating the PLAYBOOK_LIMIT for this build. (Optional)')
    ba_parser.add('--build-id', env_var='BUILD_ID',
                  help='Jenkins build ID. (Optional)')
    ba_parser.add('--build-url', env_var='BUILD_URL',
                  help="Direct-URL to the Jenkins build. (Optional)")
    ba_parser.add('--build-status', choices=['SUCCESS', 'UNSTABLE', 'FAILURE', 'NOT_BUILD', 'ABORTED'],
                  help='The build status to be reported in the comment. Accepts Jenkins build statuses.')



import sys

import ntsbuildtools.cli
import ntsbuildtools.subcommands


def main():
    ntsbuildtools.cli.set_version(ntsbuildtools.__version__)
    ntsbuildtools.cli.register(ntsbuildtools.subcommands)
    enable_argcomplete_autocomplete(ntsbuildtools.cli.get_parser())
    ntsbuildtools.cli.run(sys.argv[1:])


def enable_argcomplete_autocomplete(parser):
    """Enable command-line interface tab-autocompletion via the argcomplete package.

    If the argcomplete package is not available, this will do nothing.

    Args:
        parser (ArgumentParser): The argument parser to configure for tab-completion.
    """
    try:
        import argcomplete
        argcomplete.autocomplete(parser)
    except ImportError:
        pass


if __name__ == '__main__':
    main()



import os


class MarkdownComment:
    def __init__(self, message=None):
        """Generate a 'Markdown formatted' string. 

        Create  for various platforms that require Markdown-formatted messaging.

        Args:
            message (str, optional): The (first) message to be added to the comment. Subsequent messages can be added
            with various 'add_*' methods. Defaults to None.
        """
        self._build_annotation = []
        self._comment_data = []
        if message:
            self.add_mesage(message)

    def add_heading(self, level: int, heading: str):
        heading = f"\n{level * '#'} {heading}\n"
        self.add_message(heading, clean_vert_space=False)

    def add_message(self, message, clean_vert_space=True):
        """The core method that 'adds a message' to this MarkdownComment. 
            (Note to developers: All other MarkdownComment methods that add content to this 
            MarkdownComment *should* call this method instead of directly modifying _comment_data).

        Args:
            message (str): The message to be added.
            clean_vert_space (bool, optional): Allow MarkdownComment to clean vertical space in the message. 
                Defaults to True.

        Returns:
            MarkdownComment: Return self to allow a more fluent API -- i.e. enables chaining methods.
        """
        if clean_vert_space:
            # Remove unnecessary vertical whitespace (via rstrip)
            self._comment_data += [line.rstrip()
                                   for line in message.splitlines()]
        else:
            self._comment_data.append(message)
        return self

    def add_code_markdown(self, text, code_type=''):
        """Add a message in a 'code markdown' block. Customized to enable 'code types', which are available in
        Bitbucket.

        Args:
            text (str): The text to be put in the markdown block.
            code_type (str, optional): The 'code_type' which will be inserted in the Markdown to indicate how the code
            should be rendered. I.e. '```{style}'.

        Returns:
            MarkdownComment: Return self to allow a more fluent API -- i.e. enables chaining methods.
        """
        self.add_message(f'\n```{code_type}')
        self.add_message(text, clean_vert_space=False)
        self.add_message('\n```\n', clean_vert_space=False)
        return self

    def add_diff_markdown(self, text):
        """Add a message in a 'diff code markdown' block. Customized to enable 'code types', which are available in Bitbucket.

        Args:
            text (str): The text to be put in the markdown block.

        Returns:
            MarkdownComment: Return self to allow a more fluent API -- i.e. enables chaining methods.
        """
        self.add_code_markdown(text, code_type='diff')
        return self

    def trim_comment(self, tail, verbose=True):
        """Trim the comment down to the last 'tail' # of lines. E.g. `trim_comment(10)` means 'keep only the last 10
        lines.' NOTE: Code Markdown and any messages added with 'clean_vert_space = False' will not be trimmed --
        effectively they have a 'page break' behavior.

        Args:
            tail (int, optional): The amount of lines to be kept from the end of the comment.
        """
        if len(self._comment_data) > tail:
            self._comment_data = self._comment_data[-tail:]
            if verbose:
                self._comment_data.insert(0, '... trimmed for brevity...')

    def add_build_annotation(self, jenkins_build_id=None, jenkins_build_url=None, playbook_limit=None,
                             build_status=None):
        """Add a (Jenkins) build annotation to the build. All arguments are optional, in which case a very simple
        annotation will be prepended to the comment.

        Args:
            playbook_limit (str): The playbook_limit of this particular build, to be displayed in the build annotation
            (optional).
            jenkins_build_id (str): The build ID of the Jenkins build, to be displayed in the build annotation
            (optional). Only used if the build_url is provided as well.
            jenkins_build_url (str): The build URL of the Jenkins build, to be displayed in the build annotation
            (optional). Only used if the build_id is provided as well.

        Returns:
            MarkdownComment: Return self to allow a more fluent API -- i.e. enables chaining methods.
        """
        # Build up a string in a list -- the 'build annotation list' (aka 'bal')
        self._build_annotation.append(
            '> *This comment was automatically generated from Jenkins.* ')

        if jenkins_build_id and jenkins_build_url:
            self._build_annotation.append(
                f'> *View more details in [Jenkins build {jenkins_build_id} '
                f'(link to jenkins build, requires authorization)]({jenkins_build_url}) '
                f'[(Console Output)]({jenkins_build_url}/console).* ')

        if playbook_limit:
            # Make the playbook_limit a bit easier to read by sorting it and adding space-separation
            as_list = playbook_limit.split(',')
            as_list.sort(key=len)
            playbook_limit = ", ".join(as_list)
            playbook_limit = playbook_limit.strip()
            # Add the build annotation
            self._build_annotation.append(
                "> NOTICE: To adjust the Playbook limit from this pull request, click the '...' menu in the top right "
                "and select 'Build in Jenkins'.")
            self._build_annotation.append(
                f'\n\n Playbook limit: **{playbook_limit}**')

        # if the build status is provided, then print it in bold for all to see!
        if build_status:
            self._build_annotation.append(
                f'\n#### ***BUILD {build_status}***\n')

        # Add a horizontal line
        self._build_annotation.append("\n___\n")
        return self

    def add_text_horizontal_rule(self):
        """Add a horizontal rule (---) to the comment.

        Returns:
            MarkdownComment: Return self to allow a more fluent API -- i.e. enables chaining methods.
        """
        # vertical whitespace is trimmed nicely in Teams, so use generous newlines.
        self.add_message("\n\n___\n\n")
        return self

    def render(self):
        """'Render' this MarkdownComment as a string.

        Returns:
            str: This markdown comment, formatted.
        """
        end_of_comment_newline = ['\n']
        return os.linesep.join(self._build_annotation + self._comment_data + end_of_comment_newline)

    def __str__(self):
        """Convert this MarkdownComment into a string (overridden).

        Returns:
            str: This markdown comment, formatted.
        """
        return self.render()



import ntsbuildtools.exceptions


class TeamsCard:
    """TeamsCard represents a Microsoft Teams MessageCard (Legacy). The card generated by the to_json() method
    should be used as the HTTP POST payload to a 'Incoming Webhook Connector'.
    """

    def __init__(self):
        self.potential_actions = []
        self.sections = []
        self.title = None
        # By default, send almost-empty text (otherwise the API throws a fit)
        self.text = ' '
        self.section_index = 0

    def add_section(self, text=None):
        new_section = TeamsCardSection()
        self.sections.append(new_section)
        if text:
            new_section.add_text(text)
        return new_section

    def add_button(self, text, url):
        """[summary]

        Args:
            text ([type]): [description]
            url ([type]): [description]

        Returns:
            [type]: [description]
        """
        self.potential_actions.append(
            {
                "@type": "OpenUri",
                "name": text,
                "targets": [{
                    "os": "default",
                    "uri": url
                }]
            }
        )
        return self

    def set_text(self, text):
        # Set to non-empty string no matter what (see e2e test case: test_build_status_card_success_with_empty_file)
        self.text = text if text else ' '
        return self

    def set_title(self, title):
        self.title = title
        return self

    def to_json(self):
        return {
            "@context": "https://schema.org/extensions",
            "@type": "MessageCard",
            "themeColor": "0072C6",
            "title": self.title,
            "text": self.text,
            "sections": [section.to_json() for section in self.sections],
            "potentialAction": self.potential_actions
        }


class TeamsCardSection:
    """TeamsCardSection is a section within a Teams card. It can contain 'facts', 'text', as well many other data
    as specified in: https://docs.microsoft.com/en-us/outlook/actionable-messages/message-card-reference#section-fields
    """

    def __init__(self):
        self.facts = []
        self.text_items = []
        self.start_group = False

    def set_start_group(self):
        self.start_group = True
        return self

    def add_fact(self, label, text):
        self.facts.append({
            "name": label,
            "value": text,
        })
        return self

    def add_text(self, text: str):
        self.text_items.append(text)
        return self

    def to_json(self):
        if len(self.text_items) > 0:
            return {"start_group": self.start_group, "facts": self.facts, "text": "\n".join(self.text_items)}
        return {"start_group": self.start_group, "facts": self.facts}


class TeamsCardTemplate(TeamsCard):
    def __init__(self, required_args, optional_args, args):
        """TeamsCardTemplate is an abstract extension of the TeamsCard class that should be used as 
        the base class when creating new 'preformatted TeamsCards' classes.

        Args:
            required_args (set[str]): Set of arguments that are required to create this card.
            optional_args (set[str]): Set of arguments that are optional when creating this card.
            args (Args): The collection of arguments (implementors should pass through args from).  generally a parsearg.Namespace object. 

        A simple code example for creating a preformatted TeamsCard is below:

        ```
            class HelloWorldCard(TeamsCardTemplate):
                def __init__(self, args):
                    required_args = {'name'}
                    optional_args = {'greeting'}
                    super.__init__(required_args, optional_args, args)

                def build(self, args):
                    self.set_title(f'Hello, {args.name}!')
                    if args.greeting: # overwrite title if a optional 'greeting' argument is given.
                        self.set_title(f'{args.greeting}, {args.name}!')
        ```

        """
        super().__init__()
        self.required_args = required_args
        self.optional_args = optional_args
        # provided_args excludes arguments that have value 'None'
        all_args = vars(args).items()
        self.provided_args = set(
            [None if val is None else arg for arg, val in all_args])
        self._check_args()
        self.build(args)

    def build(self, args):
        """Factory Method that should be overwritten by classes that extend TeamsCardTemplate.

        Args:
            args: Arguments that will be used by the card-build process.
        """
        # The minimum viable card doesn't require any arguments, but it'll still take the title if given
        if hasattr(args, 'title'):
            self.set_title(args.title)
        else:
            self.set_title("[DEFAULT_CARD_TITLE]")

    def _check_args(self):
        """Check that all the arguments required by the _build method are available.

        Raises:
            ntsbuildtools.exceptions.BTTeamsCardFactoryError: If any required 
                arguments are missing.
        """
        if not self.required_args.issubset(self.provided_args):
            raise ntsbuildtools.exceptions.BTTeamsCardFactoryError(
                self._build_missing_args_message())

    # def _build_missing_args_message(self):
    #     message = ['Missing required arguments: ']
    #     # Add the list of arguments that were missing
    #     message.append(', '.join( self.required_args - self.provided_args ) + '.' )
    #     if len(self.optional_args) > 0:
    #         message.append('\n    [INFO] Optional arguments: ')
    #         message.append(', '.join( self.optional_args - self.provided_args ) + '.')
    #     return "".join(message)

    def _build_missing_args_message(self):
        my_type = type(self).__name__ 
        msg = f'Missing required arguments when creating {my_type}'
        msg = f'{msg}: {self._infer_missing_required_arguments()}.'
        # Add the list of arguments that were missing
        if self._infer_missing_optional_arguments():
            msg = f'{msg}\nOptional arguments for {my_type}'
            msg = f'{msg}: {self._infer_missing_optional_arguments()}.'
            msg = f"{msg} Was '{my_type}' the type of card you meant to build?"
            msg = f'{msg} If not, review the TeamsCardFactory.build() method for that'
            msg = f'{msg} card type to understand how this card type was chosen (based'
            msg = f'{msg} on the arguments you provided).'
        return msg

    def _infer_missing_required_arguments(self):
        return TeamsCardTemplate._infer_missing_arguments(
            self.required_args, self.provided_args)

    def _infer_missing_optional_arguments(self):
        return TeamsCardTemplate._infer_missing_arguments(
            self.optional_args, self.provided_args)

    @staticmethod
    def _infer_missing_arguments(expected, provided) -> str:
        missing = expected - provided
        if len(missing) == 0:
            return None
        missing_args = [f"--{arg.replace('_', '-')}" for arg in missing]
        return ', '.join(missing_args)



import inspect
import sys
from typing import Callable, List

import anytree
import configargparse

import ntsbuildtools.exceptions


class ArgParserNode(anytree.NodeMixin):
    """A Tree data structure for building an ArgParser with many subcommands.
    """
    def __init__(self, command: str, 
                 parser: configargparse.ArgumentParser, 
                 parent: 'ArgParserNode' = None, 
                 children: List['ArgParserNode'] = None):
        self._command = command
        self._parser = parser
        self.parent = parent
        if children:
            self.children = children
        self._subparser_factory = None

    @property
    def command(self):
        return self._command

    def _create_subparser_factory(self):
        """Creating a subparser requires tracking a 'subparser action object' -- we refer to this
        as the 'subparser_factory.' This method creates and configures that 'subparser action object'.
        """
        if sys.version_info.minor >= 7:
            self._subparser_factory = self._parser.add_subparsers(
                required=True, 
                dest="subcommand", 
                parser_class=configargparse.ArgParser
            )
        else:  # Python 3.6 support: 'required=True' argument may cause TypeError.
            self._subparser_factory = self._parser.add_subparsers(
                dest="subcommand", 
                parser_class=configargparse.ArgParser
            )
            self._subparser_factory.required = True

    @staticmethod
    def _sanitize_subcommand(subcommand: str):
        """Clean up any characters from the subcommand that does not meet our naming convention.

        Args:
            subcommand (str): The subcommand to be cleaned/checked.

        Returns:
            str: The subcommand without any invalid characters
        """
        return subcommand.replace('_','-')

    def _add_subparser(self, subcommand, help_msg):
        """Add a subparser to this node's (private) ArgumentParser, and return it. This
        is a wrapper to simplify interfacing with ArgumentParser's `add_subparsers` method.

        Args:
            subcommand (str): The name of the subcommand/subparser to be added.
            help_msg (str): The help message 

        Returns:
            configargparse.ArgumentParser: The subparser!
        """
        if self._subparser_factory is None:
            self._create_subparser_factory()
        return self._subparser_factory.add_parser(subcommand, description=help_msg)

    def add(self, subcommand: str, help_msg: str) -> 'ArgParserNode':
        """Add a subcommand directly to this Node. This is generally used for adding intermediate
        subcommands that do not have a `default_func` associated with them.

        Args:
            subcommand (str): The name of the subcommand to be added to the CLI. Any underscores 
                              will be replaced with hyphens.
            help_msg (str): The help message that will show in the CLI for this subcommand.

        Returns:
            ArgParserNode: The added node.

        Example:
            Adding an inner 'example' CLI path: "buildtools example". This is the first step
            to creating a subcommand with an actual default_func, where the functionality
            lies. See `ArgParserNode.add_leaf()` for more information.


                Setup the "Root ArgParser" and the Tree data structure.
                >>> _root_parser = configargparse.ArgumentParser()
                >>> tree_root = ArgParserNode(command='buildtools', parser=_root_parser)

                Add the 'example' subcommand node!
                >>> tree_root.add('example', 'Just some example subcommands.')
                ArgParserNode(command='example')
            """
        subcommand = ArgParserNode._sanitize_subcommand(subcommand)
        new_child_parser = self._add_subparser(subcommand, help_msg)
        child_node = ArgParserNode(parser=new_child_parser, command=subcommand, parent=self)
        return child_node

    def add_leaf(self, subcommand: str, help_msg: str, default_func: Callable[[configargparse.Namespace],None], 
                 config_parser: Callable[[configargparse.ArgumentParser],None] = None) -> 'ArgParserNode':
        """Add a leaf node -- the leaf *must* provide a `default_func` Callable. 
        Additionally, the `config_parser` Callable may be provided, e.g. if your default_func 
        has any parameters (see the example).

        The leaf nodes are where the true functionality of the CLI is provided. The leaf nodes
        hold onto the 'default functions' which are the actual 'subcommands' provided in the
        `subcommands` package.

        Implementation Note: The `default_func` is assigned using a pattern discussed in `argparse` documentation.


        Args:
            subcommand (str): The name of the leaf-subcommand to be added. Any underscores 
                              will be replaced with hyphens.
            help_msg (str): The help message that will show in the CLI for this subcommand.
            default_func (Callable): The default function -- effectively the 'main' function 
                that corresponds to this subcommand.
            config_parser (Callable[[configargparse.ArgumentParser], None]): If the default 
                function requires any parameters, those should be configured via this  
                Callable (see the example).

        Returns:
            ArgParserNode: The added node.

        Example:
            Adding a "Hello, {Your name here}!" functionality. This has the CLI arguments 
            like: "hello <name>".

                Setup the "Root ArgParser" and the Tree data structure.
                >>> _root_parser = configargparse.ArgumentParser()
                >>> _tree_root = ArgParserNode(command='buildtools', parser=_root_parser)

                Add the 'hello' subcommand -- including Callables!
                >>> def config_parser_for_hello_function(parser):
                ...    parser.add_argument('name')
                >>> def hello_function(args):
                ...    print(f"Hello, {args.name}!")
                >>> _tree_root.add_leaf('hello', "Just some basic 'hello, world!' functionality", 
                ...     hello_function, config_parser_for_hello_function)
                ArgParserNode(command='hello')

                Parse arguments
                >>> parsed_args = _root_parser.parse_args(['hello', 'Fred'])

                This will execute `hello_function` (with args.name = 'Fred')
                >>> parsed_args.func(parsed_args)
                Hello, Fred!
        """
        child_node = self.add(subcommand, help_msg)
        child_node.set_default_func(default_func)
        if config_parser:
            child_node.config_parser(config_parser)
        return child_node

    def config_parser(self, config_parser_func: Callable[[configargparse.ArgumentParser], None]):
        """Configure the parser that is stored in this node.

        Args:
            config_parser (Callable): A callable that will configure this ArgParserNode's internal ArgumentParser.
        """
        if list(inspect.signature(config_parser_func).parameters) != ["parser"]:
            raise ntsbuildtools.exceptions.BTConfigParserSignatureError(config_parser_func.__module__)
        config_parser_func(self._parser)

    def set_default_func(self, default_func: Callable):
        """Provide the 'main' functionality of this particular ArgParserNode.

        Args:
            default_func (Callable): The default function for this ArgParser -- effectively the 
                'main' function that corresponds to this subcommand.

        Raises:
            BTMainFuncSignatureError: If the default_func parameters do not match our convention.
        """
        if list(inspect.signature(default_func).parameters) != ["args"]:
            raise ntsbuildtools.exceptions.BTMainFuncSignatureError(default_func.__module__)
        self._parser.set_defaults(func=default_func)

    def __repr__(self):
        return f"ArgParserNode(command='{self._command}')"



"""Simple Hello World functionality."""
from configargparse import ArgParser

def config_parser(parser: ArgParser):
    parser.add_argument('--verbose', '-v', action='store_true', 
                        help="Be extra verbose when saying 'hello world!'")

def main(args):
    if args.verbose:
        print("With all the verbosity in the world... Hello World!")
    else:
        print("Hello World!")



#!/usr/bin/env python3
"""This module provides the `register` function for establishing the command-line interface,
and the `run` function to act as the main entry point for the entirety of ntsbuildtools.
"""
from ntsbuildtools.cli.framework import run, register, get_parser, set_version


__all__ = [ 'run', 'register', 'get_parser', 'set_version' ]



import re

def _last_exception_message(exception_message):
    """Extract just the LAST message from the provided exception stack trace.
    """
    try:
        lines = str.splitlines(exception_message)
        message = []
        for line in reversed(lines):
            if re.search(r'line \d+, in', line):
                break
            message.insert(0, line)
        return "\n".join(message)
    except TypeError:
        return str(exception_message)


class ReturnValues:
    def __init__(self, return_values_raw_data):
        """Represents the Ansible ReturnValues that are produced when an Ansible Task 
        is executed against a targetted host.

        This data structure corresponds to the 'common' 
        https://docs.ansible.com/ansible/latest/reference_appendices/common_return_values.html#common

        Args:
            return_values_raw_data (Dict): The raw 'ReturnValues' data.
        """
        self._data = return_values_raw_data

    @property
    def data(self):
        return self._data

    @property
    def failed(self) -> bool:
        if 'failed' not in self.data:
            return False
        # Convert to a bool -- e.g. `failed` might be a string or an object! 
        return True if self.data['failed'] else False

    @property
    def changed(self) -> bool:
        if 'changed' not in self.data:
            return False
        # Convert to a bool -- e.g. `changed` might be a string or an object! 
        return True if self.data['changed'] else False

    def pretty_error(self):
        if 'results' in self.data:
            for return_values in self.data['results']:
                # TODO: How to use the factory here instead of directly building the devault ReturnValues? (circular dependency issue)
                # TODO: Should we accumulate child-failures before returning?
                child = ReturnValues(return_values)
                if child.failed:
                    return f"(loop_item='{child.data['item']}') {child.pretty_error()}"
        if 'error' in self.data:
            return str(self.data['error'])
        if 'exception' in self.data:
            return _last_exception_message(self.data['exception'])
        if 'msg' in self.data:
            return str(self.data['msg'])
        return "[ntsbuildtools] Unable to parse error."

    def pretty_diff(self):
        if not self.changed:
            return "[ntsbuildtools] No changes."
        if 'diff' in self.data:
            return str(self.data['diff'])
        if 'changed' in self.data:
            return f"[ntsbuildtools] Changed status: {str(self.data['changed'])}"
        return "[ntsbuildtools] Unable to parse diff."



"""Infer the `--limit` for a call to ansible-playbook.

Given an Ansible inventory file (hosts, ini format) and a list of 'changed group/host_vars files'
infer an appropriate `--limit` to be used in a call to `ansible-playbook`.
"""
import configargparse

from ntsbuildtools.ansible import InventoryTree

# __doc__ = __doc__ + InventoryTree.infer_playbook_limit.__doc__


def config_parser(parser: configargparse.ArgParser):
    parser.add_argument(
        'source',
        help='List of all paths of an Ansible Inventory that have been changed.'
    )
    parser.add_argument(
        'inventory',
        help='The path to the relevant (`ini` styled) Ansible hosts file.'
    )

def main(args: configargparse.Namespace):
    # Simplify the playbook limit
    playbook_limit = infer_playbook_limit(args.source, args.inventory)

    # Print the simplified playbook limit to stdout
    print(",".join(list(playbook_limit)))


def infer_playbook_limit(source_path, inventory_path):
    inventory = InventoryTree(inventory_path)
    with open(source_path) as changed_files:
        return inventory.infer_playbook_limit(changed_files.readlines())



'''Posts a comment to a GitHub pull request.

If a comment is too large to fit as a single comment in github, this solution can 
fragment a comment into pieces (via the `--max-comment-size` argument).
'''
import mdfrag
import requests

import ntsbuildtools.io
from ntsbuildtools.markdown.comments import MarkdownComment


def config_parser(parser):
    # Parse the 'comment type' with mutex logic -- don't allow user to provide more than one type of message from the provided options.
    message_parser = parser.add_mutually_exclusive_group(required=True)
    message_parser.add('--message', '-m', dest='comment_message',
                       help='Provide the comment as a message on the command line.')
    message_parser.add('--file', '-f', dest='comment_file',
                       help='Provide the comment in a file.')
    # Core parsing bits
    parser.add('--user', required=True, env_var='BITBUCKET_USER',
               help="GitHub user that will be used to authenticate to GitHub.")
    parser.add('--password', required=True, env_var='BITBUCKET_PASSWORD',
               help="GitHub password (or Personal Access Token) for the GitHub user.")
    parser.add('--project', required=True, env_var='GITHUB_PROJECT',
               help='The GitHub "owner" or "organization" for the repository.')
    parser.add('--repo', required=True, env_var='GITHUB_REPO',
               help='The GitHub repository slug for the repository.')
    parser.add('--pull-request-id', required=True, env_var='PR_ID',
               help='The ID of the GitHub pull request to be commented on.')
    # Group of arguments to do with formatting.
    formatting = parser.add_argument_group('formatting')
    formatting.add('--max-comment-size', type=int,
                   help='Fragment the comment into based on the maximum comment size.')
    ba_parser = formatting.add_argument_group('Build Annotation')
    ba_parser.add('--build-annotation', action='store_true',
                  help='Provide a "build annotation" in the comment, with information about a Jenkins build.')
    ba_parser.add('--playbook-limit', env_var='PLAYBOOK_LIMIT',
                  help='A string indicating the PLAYBOOK_LIMIT for this build. (Optional)')
    ba_parser.add('--build-id', env_var='BUILD_ID',
                  help='Jenkins build ID. (Optional)')
    ba_parser.add('--build-url', env_var='BUILD_URL',
                  help="Direct-URL to the Jenkins build. (Optional)")
    ba_parser.add('--build-status', choices=['SUCCESS', 'UNSTABLE', 'FAILURE', 'NOT_BUILD', 'ABORTED'],
                  help='The build status to be reported in the comment. Accepts Jenkins build statuses.')

def main(args):
    # Get the core comment message from args -- these options are mutually exclusive (see config_parser)
    comment = MarkdownComment()
    if args.comment_message:
        comment.add_message(args.comment_message)
    elif args.comment_file:
        message_data = ntsbuildtools.io.readfile(args.comment_file)
        comment.add_message(message_data)
    if args.build_annotation:
        comment.add_build_annotation(args.build_id, args.build_url, args.playbook_limit, args.build_status)

    if len(str(comment).strip()) == 0:
        raise ValueError(
            "Some comment contents must be provided -- the comment to be sent was an empty string.")

    # Actually make the HTTP Post request to publish the comment
    try:
        post_url = f'https://api.github.com/repos/{args.project}/{args.repo}/issues/{args.pull_request_id}/comments'
        # If the comment will exceed the max comment size, do fragmentation instead!
        if args.max_comment_size and len(str(comment)) > args.max_comment_size:
            _split_and_post(post_url, args.user, args.password, comment, args.max_comment_size)
        else:
            _post_comment(post_url, args.user, args.password, comment)
    except requests.exceptions.HTTPError as error:
        print(f"[ERROR] HTTP Error ({error.response.status_code}): {error.response.text}")
        exit(-1)

def _post_comment(url, user, password, comment):
    response = requests.post(
        url, 
        json={'body': str(comment) if comment else '*Empty string provided.*'},
        auth=(user, password)
    )
    response.raise_for_status()

def _split_and_post(url, user, password, comment, max_comment_size):
    comment_fragments = mdfrag.split_markdown(str(comment), max_comment_size)
    # Comments are in reverse chronological, so we should render the fragments in reverse.
    comment_fragments.reverse()
    # cp_i => "Comment Part Index"
    cp_i = len(comment_fragments)
    for comment_part in comment_fragments:
        pagination_str = f'> Comment __{cp_i}__ of __{len(comment_fragments)}__.'
        msg = f'{pagination_str}\n{comment_part}'
        cp_i -= 1
        response = requests.post(url, json={'text': msg}, auth=(user, password))
        response.raise_for_status()



"""Set the "Build Status" in Bitbucket."""
import requests
import configargparse 
import os

def config_parser(parser: configargparse.ArgParser):
    parser.add('--bitbucket-url', env_var='BITBUCKET_URL', help="URL for Bitbucket.")
    parser.add('--commit', required=True, env_var='COMMIT',
               help='The FULL LENGTH git commit for which this build is being run.')
    parser.add('--build-status', required=True,
               choices=['STARTED', 'SUCCESS', 'UNSTABLE', 'FAILURE', 'NOT_BUILD', 'ABORTED'],
               help='The build status to be reported to Bitbucket. Accepts Jenkins build statuses, as well as the keyword "STARTED".')
    parser.add('--user', required=True, env_var='BITBUCKET_USER',
               help="Bitbucket user that will be used to authenticate to Bitbucket.")
    parser.add('--password', required=True, env_var='BITBUCKET_PASSWORD',
               help="Bitbucket password (or Personal Access Token) for the Bitbucket user.")
    parser.add('--build-url', required=True, env_var='BUILD_URL',
               help="Direct-URL to the Jenkins build.")
    parser.add('--job-name', required=True, env_var='JOB_NAME',
               help="The Jenkins Job/Pipeline name that called this build.")
    parser.add('--build-key', env_var='BUILD_KEY', default="UPDATE_ONLY",
               help="""Unique identifier to differentiate consecutive builds (optional). If omitted, 
        consecutive builds will *update* the existing build status for this Jenkins Job/Pipeline.""")
    parser.add('--digital-ocean', action='store_true',
               help="""Use the Digital Ocean front end for build URL links""")
    parser.add('-m', '--message', help="Build message, to be shown in Bitbucket. Should be kept very short.")


def _parse_status(jenkins_status):
    if jenkins_status == "STARTED":
        return "INPROGRESS"
    if jenkins_status == "SUCCESS":
        return "SUCCESSFUL"
    # By default, fail the build (and for all the other build statuses)
    else:
        return "FAILED"


def _build_json(status, build_key, job_name, build_url, digital_ocean, message=None):
    status = _parse_status(status)
    # override build_url if digital_ocean is true.
    if digital_ocean:
        # we don't use the env_var in config_parser because it'll always be true if digital
        # ocean is installed. This way allows the user to choose.
        build_url = os.getenv('RUN_DISPLAY_URL')
    if message:
        return {
            "state": status,
            "key": build_key,
            "name": job_name,
            "url": build_url,
            "description": message
        }
    return {
        "state": status,
        "key": build_key,
        "name": job_name,
        "url": build_url
    }


def main(args):
    json_payload = _build_json(args.build_status, args.build_key,
                               args.job_name, args.build_url, args.digital_ocean, args.message)
    try:
        response = requests.post(f'{args.bitbucket_url}/rest/build-status/1.0/commits/{args.commit}',
                                 json=json_payload, auth=(args.user, args.password))
        response.raise_for_status()
    except requests.exceptions.HTTPError as error:
        print(f"[ERROR] HTTP Error ({error.response.status_code}): {error.response.text}")
        exit(-1)



"""util.py -- Utility module for the buildtools project."""


def is_nonempty_str(obj):
    return isinstance(obj, str) and len(obj) > 0


def hasattr_nonempty_str(obj, attribute):
    return hasattr(obj, attribute) and is_nonempty_str(getattr(obj, attribute))



"""All of the custom Errors that BuildTools 
"""
from textwrap import dedent

class BTError(Exception):
    """The base class for all BuildTools errors."""
    pass


class BTConventionError(BTError):
    """Errors that will be thrown if the BuildTools `subcommands` package breaks 
    any of the conventions outlined in our "CLI Framework" documentation. 
    """
    pass


class BTMainFuncMissingError(BTConventionError):
    def __init__(self, module_name):
        message = dedent(f"""\
            module {module_name} is missing its 'main' method! 
            Each module in the subcommands package must define a 'main(args)' function.
            """)
        super().__init__(message)


class BTConfigParserSignatureError(BTConventionError):
    def __init__(self, module_name):
        message = dedent(f"""\
            module {module_name} has the incorrect arguments for its 'config_parser' function! 
            The signature should be `def config_parser(parser: configargparse.ArgParser)`.
            """)
        super().__init__(message)


class BTMainFuncSignatureError(BTConventionError):
    def __init__(self, module_name):
        message = dedent(f"""\
            module {module_name} has the incorrect arguments for its 'main' function! 
            The signature should be `def main(args: configargparse.Namespace)`.
            """)
        super().__init__(message)


class BTTeamsCardFactoryError(BTError):
    pass



#!/usr/bin/env python3
"""This module provides the `register` function for establishing the command-line interface,
and the `run` function to act as the main entry point for the entirety of ntsbuildtools.
"""
import configargparse
import importlib
import os
import pkgutil

import ntsbuildtools.exceptions
import ntsbuildtools
from ntsbuildtools.cli.tree import ArgParserNode


# Initialize the "Root ArgumentParser" for the whole project
_root_parser = configargparse.ArgumentParser(
    description="Tools that help with Jenkins build scripts.", 
    default_config_files=["./buildtools.yaml", "./buildtools.ini"]
)

# Intitialize the ArgParserNode (also for the whole project)
_parser_tree = ArgParserNode(command='buildtools', parser=_root_parser)

def set_version(version: str):
    global _root_parser
    _root_parser.add_argument('--version', action='version', 
                              version=f"BuildTools CLI {version}")


def register(package) -> ArgParserNode:
    """Register all the modules found in the package. `package` will be fully traversed to
    collect the 'docstrings' for each subcommand in the CLI tree. Additionally, the `main` 
    and `config_parser` methods for each subcommand are processed into the root parser.

    Once registration of subcommands is complete, invoking `run(sys.argv[:1])` function
    will execute the appropriate 'main' function depending on the command-line arguments provided.

    Args:
        package (package|str): The `subcommands` package to be registered (or name of the package).

    Returns:
        ArgParserNode: The CLI tree data structure, which now contains the registered modules.

    Example:
        The simplest example is also representative of the 
            import ntsbuildtools.cli
            import ntsbuildtools.subcommands

            ntsbuildtools.cli.register(ntsbuildtools.subcommands)
            ntsbuildtools.cli.run(sys.argv[1:])
    """
    if isinstance(package, str):
        package = importlib.import_module(package)
    _register(package, _parser_tree)
    return _parser_tree


def children_module_info_of(package):
    if isinstance(package, str):
        package = importlib.import_module(package)
    path_to_package = os.path.dirname(package.__file__)
    return [_ for _ in pkgutil.iter_modules([path_to_package])]


def _register(package, node):
    for module_info in children_module_info_of(package):
        module_name = f'{package.__name__}.{module_info.name}'
        if module_info.ispkg:
            # If this is a package (think directory), create an inner node and RECURSE!
            loaded_package = importlib.import_module(module_name)
            child = node.add(module_info.name, loaded_package.__doc__)
            _register(loaded_package, child)
        else:
            # If this is a module (think .py file), create a leaf node!
            loaded_module = importlib.import_module(module_name)
            if not hasattr(loaded_module, 'main'):
                raise ntsbuildtools.exceptions.BTMainFuncMissingError(module_name)
            leaf = node.add_leaf(module_info.name, loaded_module.__doc__, loaded_module.main)
            # Each subcommand can OPTIONALLY provide a config_parser(parser) function
            if hasattr(loaded_module, 'config_parser'):
                leaf.config_parser(loaded_module.config_parser)


def get_parser():
    return _root_parser


def run(args):
    """Run BuildTools!"""
    parsed_args = _root_parser.parse_args(args)
    parsed_args.func(parsed_args)



import itertools
import logging
import os
from configparser import ConfigParser, NoSectionError
from dataclasses import dataclass, field
from functools import reduce
from typing import List, Set

logger = logging.getLogger(__name__)


def flatten(lst):
    return [item for sublist in lst for item in sublist]


@dataclass
class InventoryNode:
    """Represents a host or group of hosts.

    Has a name and parent/children relationships.
    """
    name: str
    children: Set['InventoryNode'] = field(default_factory=set)
    parents: Set['InventoryNode'] = field(default_factory=set)

    def __hash__(self):
        return hash(self.name)

    def add_child(self, child):
        """Add a child to this nodes children."""
        self.children.add(child)

    def add_parent(self, parent):
        """Add a parent to this nodes parents."""
        self.parents.add(parent)

    def get_ancestors(self) -> Set['InventoryNode']:
        """Return the set of ancestors of this node.
        """
        if self.parents:
            ancestors = reduce(Set.union, map(InventoryNode.get_ancestors, self.parents))
            return self.parents.union(ancestors)
        else:
            return set()

    def get_younglings(self):
        if self.children:
            younglings = reduce(Set.union, map(InventoryNode.get_younglings, self.children))
            return self.children.union(younglings)
        else:
            return set()

    def __str__(self):
        return self.name

    def __repr__(self):
        """Creates a useful representation for debugging.

        Returns:
            str: This object with parents and children listed as strings.
        """
        children_names = [child.name for child in self.children]
        parent_names = [parent.name for parent in self.parents]
        return f'InventoryNode(name={self.name}, children=set({children_names}), parents=set({parent_names}))'


def strip_ansible_group_name_metadata(group_name):
    return group_name.replace(':children', '') 


class InventoryTree:
    def __init__(self, path: str):
        """Ansible Inventory.

        Enables traversing and querying an Ansible Inventory (e.g. your ini-style "hosts" file).

        Args:
            path (str): Path to the Ansible Inventory config file, your ini-style `hosts` file.
        """
        self.nodes = set()
        self.root = set()
        config = ConfigParser(allow_no_value=True)
        config.read(path)
        self._init_nodes(config)
        self._init_nodes_relationships(config)
        self._init_root()

    def get_node(self, name: str) -> InventoryNode:
        """Find node in this tree.

        Does not ensure the node name is unique -- will return first node with name. 
        """
        name = strip_ansible_group_name_metadata(name)
        for node in self.nodes:
            if node.name == name:
                return node
        raise LookupError(f'Unable to find Ansible Inventory item with name: {name}')

    def infer_playbook_limit(self, files_touched: List[str], minimize=True) -> List[str]:
        """Infer an Ansible 'Playbook Limit' from a list of 'touched [host|group]_vars files.'

        This can be very useful for Continuous Integration (CI) of Ansible playbooks.

        Args:
            files_touched (List[str]): List of Ansible inventory files changed (only care 
                changes to host_vars and group_vars). 
            minimize (bool, optional): Remove any groups that are redundant. Defaults to True.

        Example:
            Imagine using Git to track an Ansible project. Your goal is to determine the 
            `--limit` argument for a call to `ansible-playbook` in your CI pipeline. 

            If using GitHub Flow for the project, you can a list of files_touched 
            like the following:

            ```
            git diff origin/main --name-only --ignore-all-space \
                | grep -E "group_vars|host_vars" > files_touched.txt
            ```

            This method can take the files_touched.txt as input like the following, 

                >>> inventory = InventoryTree('tests/testresources/ansible-hosts-example')
                >>> with open('tests/testresources/files_touched.txt') as files_touched:
                ...     inventory.infer_playbook_limit(files_touched.readlines())
                ['dc-leafs', 'dr-gws']

            The produced list can be used as the `--limit` argument for a call to `ansible-playbook`.

        Returns:
            List[str]: Simplified list of groups/hosts from this Ansible Inventory, to be used as playbook_limit.
        """
        limits = []
        for file_path in files_touched:
            split_file_path = file_path.split(os.path.sep)
            if split_file_path[-2] == 'group_vars' or split_file_path[-2] == 'host_vars':
                limits.append(split_file_path[-1].strip())
            else:
                logger.warning(f'Encountered non-inventory file: {file_path}')
        return self.simplify_playbook_limit(limits, minimize)

    def simplify_playbook_limit(self, hosts_groups_touched: List[str], minimize=True) -> List[str]:
        """Simplify an Ansible 'Playbook Limit' based on a list of 'touched hosts/groups'
        """
        nodes = set([  self.get_node(name) for name in hosts_groups_touched ])

        removals = set()
        additions = set()
        first_run = True
        while removals or additions or first_run:
            first_run = False

            # 0. Process changes, removals take precedence
            nodes = nodes.union(additions)
            additions.clear()
            nodes -= removals
            removals.clear()

            #
            # 1. Remove any hosts/groups if they are already represented by another limit
            #   If the playbook limit includes one of my ancestors, then I am already included in the limits via that group!
            #
            for node in nodes:
                if node.get_ancestors().intersection(nodes):
                    removals.add(node)

            #
            # 2. Check if the playbook limit includes ALL the children of some group
            #   Add the group and remove ALL the children
            #
            parents = set().union(*[ node.parents for node in nodes ])
            parents = parents.difference(nodes)
            for parent in parents: 
                if parent.children.issubset(nodes):
                    removals = removals.union(parent.children)
                    additions.add(parent)

            #
            # 3. Finally, Check if any groups are redundant
            #
            if minimize and not removals and not additions:
                for nodeA, nodeB in itertools.combinations(nodes, 2):
                    if nodeA.children and nodeA.children.issubset(nodeB.get_younglings()):
                        removals.add(nodeA)

        return sorted([limit.name for limit in nodes])

    def _init_nodes(self, config: ConfigParser):
        """Add all hosts and groups to the tree."""
        for group_name in config.sections():
            # if group_name.endswith(':vars'):
            #     continue
            self.nodes.add(InventoryNode(strip_ansible_group_name_metadata(group_name)))
            for host_or_group_name in config.options(group_name):
                self.nodes.add(InventoryNode(host_or_group_name))

    def _init_nodes_relationships(self, config: ConfigParser):
        """Establish parent-child relationships within this tree.

        Expects all the nodes to already be initialized.
        """
        groups = [self.get_node(group_name) for group_name in config.sections()]
        for group in groups:
            try:
                group_member_names = config.options(group.name)
            except NoSectionError:
                group_member_names = config.options(f'{group.name}:children')
            for host_or_group_name in group_member_names:
                host_or_group = self.get_node(host_or_group_name)
                group.add_child(host_or_group)
                host_or_group.add_parent(group)

    def _init_root(self):
        """Establish the root set of this tree.

        TODO: Discuss what this really means.

        Expects parent-child relationships to already be established within this tree.
        """
        for node in self.nodes:
            if not node.parents:
                self.root.add(node)



from . import default

class JunosConfigReturnValues(default.ReturnValues):
    def pretty_diff(self):
        try:
            return self.data['diff']['prepared']
        except LookupError:
            return super().pretty_diff()

    def pretty_error(self):
        # NOTE: Currently this is the same as the default ReturnValues!
        return super().pretty_error()



from typing import Dict
from . import default
from . import junos_config


class ReturnValuesFactory:
    def __init__(self) -> None:
        """Create the appropriately-typed ReturnValues object, based on the Ansible
        module name. By default, will return the base class, ReturnValues.
        """
        self._constructors: Dict = {}
        self._default_constructor = default.ReturnValues

    def register(self, module_name, return_values_cls):
        """Register a ReturnValues extension Class to this Factory.

        Args:
            module_name (str): The particular Ansible Module that this ReturnValues is designed to process.
            return_values_cls (ReturnValues): A class that extends ReturnValues (i.e. and provides custom
            `pretty_diff` and `pretty_error` methods).

        Raises:
            ValueError: [description]
        """
        if module_name in self._constructors:
            raise ValueError(f"Conflicting module_name already registered: {self._constructors[module_name]}")
        if not issubclass(return_values_cls, self._default_constructor):
            raise ValueError(f"Must register a valid ReturnValues object that extends from {self._default_constructor}")
        self._constructors[module_name] = return_values_cls

    def build(self, return_values_raw, module_name):
        """Given raw Return Values from Ansible, parse them into a ReturnValues object.

        The caller can then use the `pretty_error` or `pretty_diff` methods on the returned object.

        Args:
            return_values_raw (Dict): The JSON data representing the Ansible Return Values
            module_name (str): The Ansible Module that produced these return values.

        Returns:
            ReturnValues: An object that provides `pretty_error` and `pretty_diff` methods.
        """
        try:
            return self._constructors[module_name](return_values_raw)
        except LookupError:
            return self._default_constructor(return_values_raw)


_factory = ReturnValuesFactory()
_factory.register('junos_config', junos_config.JunosConfigReturnValues)

def build(return_values_raw, module_name):
    return _factory.build(return_values_raw, module_name)



"""Parses Ansible JSON output and turns it into pretty Markdown output.

Reports on any failures. Optionally reports the 'diffs' for any number of 'crucial tasks'"""
from typing import List
import configargparse

import ntsbuildtools.io
from ntsbuildtools.markdown.comments import MarkdownComment
from ntsbuildtools.ansible.playbook.results import PlaybookResults


def main(args):
    # Parse the json output into a PlaybookResults object
    json_output = ntsbuildtools.io.readfile(args.src)
    results = PlaybookResults.from_json(json_output)

    markdown = render_playbook_result(
        results, args.crucial_task, args.crucial_task_type)

    with open(args.dst, 'w') as dst_file:
        dst_file.write(markdown)


def render_playbook_result(results: PlaybookResults,
                           crucial_tasks: List[str] = None,
                           crucial_ansible_modules: List[str] = None) -> str:
    markdown = MarkdownComment()
    for task_name in results.tasks:
        for task_result in results.by_task_name(task_name):
            if task_result.failed:
                markdown.add_message(task_result.as_markdown())
            elif crucial_tasks and task_result.task_name in crucial_tasks:
                markdown.add_message(task_result.as_markdown())
            elif crucial_ansible_modules and task_result.ansible_module in crucial_ansible_modules:
                markdown.add_message(task_result.as_markdown())
    return str(markdown)


def config_parser(parser: configargparse.ArgParser):
    parser.add_argument(
        'src',
        help='''Path to a Ansible JSON output file. This JSON file should be generated 
                using the "ansible.posix.json" Callback Plugin. (https://docs.ansible.com/ansible/latest/collections/ansible/posix/json_callback.html)'''
    )
    parser.add_argument(
        'dst',
        help='''Path to a file where the output should be saved. 
                NOTE: If the file exists, it's contents will be overwritten.'''
    )
    parser.add_argument(
        '--crucial-task', action='append',
        help="Any particular tasks that should ALWAYS be reported on. Provide the full name of that task."
    )
    parser.add_argument(
        '--crucial-task-type', action='append',
        help="""Often there will be one particular type of task that you want to see any results from. If there are any particular Ansible Modules that you'd like to see the results of, provide  type of task If there is one type of task that we should ALWAYS report on, provide its name here."""
    )



import os
import sys
from typing import List

from setuptools import find_packages, setup


# From https://packaging.python.org/guides/single-sourcing-package-version/#single-sourcing-the-package-version
def read_project_file(relative_file_path: str):
    here = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(here, relative_file_path), 'r') as file_pointer:
        return file_pointer.read()


def installation_requirements() -> List[str]:
    install_requires = [
        "requests>=2.0",
        "ConfigArgParse>=1.0,<2",
        "anytree>=2.0",
        "art>=2.0",
        "mdfrag==0.1.0",
    ]
    if sys.version_info.minor < 7:
        install_requires.append('dataclasses>=0.6')
    return install_requires


setup(
    name="ntsbuildtools",
    package_dir={'': 'src'},
    packages=find_packages('src'),
    version='4.0.5',  # We attempt to follow 'semantic versioning', i.e. https://semver.org/ 
    license='MIT',
    description="CLI toolset that supports CICD processes.",
    long_description_content_type='text/markdown',
    long_description=read_project_file('docs/user-guide.md'),
    author='Network & Telecom Svcs (University of Oregon)',
    author_email='rleonar7@uoregon.edu',
    url='https://git.uoregon.edu/projects/ISN/repos/buildtools/browse',
    keywords=['Jenkins', 'NTS', 'UO', 'CLI', 'Integrations', 'API'],
    entry_points={
        'console_scripts': [
            'buildtools=ntsbuildtools.main:main'
        ]
    },
    install_requires=installation_requirements(),
    classifiers=[  # Classifiers selected from https://pypi.org/classifiers/
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers', 
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ]
)



"""Create a Bitbucket Pull Request. (Not Implemented)"""
import os
import datetime
import requests

def main(args):
    raise NotImplementedError("This is only half implemented")


def build_pull_request_json(project, repo_slug, from_branch, to_branch, title, description):
    return {
        "title": title,
        "description": description,
        "state": "OPEN",
        "open": True,
        "closed": False,
        "fromRef": {
            "id": f"refs/heads/{from_branch}",
            "repository": {
                "slug": repo_slug,
                "project": {
                    "key": project,
                }
            }
        },
        "toRef": {
            "id": f"refs/heads/{to_branch}",
            "repository": {
                "slug": repo_slug,
                "project": {
                    "key": project
                }
            }
        },
        "locked": False
    }
    # END build_pull_request_json

def _post_pull_request(args):
    # Parameters from environment variables (should be populated by the Jenkinsfile)
    try:
        from_branch_name = os.environ['FROM_BRANCH']
        to_branch_name = os.environ['TO_BRANCH']
        git_uo_user = os.environ['USER']
        git_uo_password = os.environ['PASSWORD']
        project = os.environ['PROJECT']
        repo_slug = os.environ['REPO']
    except KeyError as internal_error:
        raise ValueError('Missing required environment variable(s), see internal exception. '
                         'NOTE: more environment variables may be missing, please consult the source code.') \
            from internal_error

    # Create the pull request JSON object to be POSTed
    pull_request_json = build_pull_request_json(
        project,
        repo_slug,
        from_branch_name,
        to_branch_name,
        title=f"""[AUTO] pip updates, {datetime.now().strftime("%d %B, %Y")}""",
        description="Automatically generated pull request to check for updates."
    )

    # Make the request
    try:
        response = requests.post(
            f'{args.bitbucket_url}/rest/api/1.0/projects/{project}/repos/{repo_slug}/pull-requests',
            json=pull_request_json, auth=(git_uo_user, git_uo_password))
        response.raise_for_status()
    except requests.exceptions.HTTPError as error:
        print(f"[ERROR] HTTP Error ({error.response.status_code}): {error.response.text}")
        exit(-1)



import enum

import ntsbuildtools.io
from ntsbuildtools.teams.card_template import TeamsCardTemplate
from ntsbuildtools.markdown.comments import MarkdownComment
from ntsbuildtools.util import hasattr_nonempty_str

class TeamsMessageTypes(enum.Enum):
    PULL_REQUEST = 'pull-request'
    BUILD_STATUS = 'build-status'

    @classmethod
    def values(cls):
        return [member.value for member in cls]

class PullRequestCard(TeamsCardTemplate):
    def __init__(self, args):
        required_arguments = {'bitbucket_url', 'project', 'repo_slug', 'pull_request_id', 'pull_request_author', 'pull_request_title', 'pull_request_dest'}
        optional_arguments = {'pull_request_description', 'playbook_limit', 'jenkins_build_url', 'build_status'}
        super().__init__(required_arguments, optional_arguments, args)

    def build(self, args):
        # Calculate URLs for the pull request, based on args
        pull_request_url = f'{args.bitbucket_url}/projects/{args.project}/repos/{args.repo_slug}/pull-requests/{args.pull_request_id}'
        pull_request_diff_url = pull_request_url + "/diff"

        # Set the title using a (module-level) helper method
        self.set_title(f"Pull Request {args.pull_request_id}: {args.pull_request_title}")
        # Section with pull request facts
        pr_facts = self.add_section()
        pr_facts.add_fact("Requested by", args.pull_request_author)
        if hasattr_nonempty_str(args, 'playbook_limit'):  # playbook_limit is optional
            pretty_playbook_limit = args.playbook_limit.replace(',', ' ︙ ')
            pr_facts.add_fact("Playbook limit", pretty_playbook_limit)
        if hasattr_nonempty_str(args, 'build_status'):  # build_status is optional
            if args.build_status == 'FAILURE':  # Add emphasis if this is a failure.
                pr_facts.add_fact("Build Status", f'***{args.build_status}***')
            else:
                pr_facts.add_fact("Build Status", args.build_status)

        pr_facts.add_fact("Destination branch", args.pull_request_dest)

        # Section with the pull request description (optional)
        if hasattr_nonempty_str(args, 'pull_request_description'): 
            self.add_section().add_text(args.pull_request_description)

        # Actions that you should take now (buttons at the bottom)!
        self.add_button(f"Review PR #{args.pull_request_id}", pull_request_url)
        self.add_button("See proposed changes!", pull_request_diff_url)
        if hasattr_nonempty_str(args, 'jenkins_build_url'):
            self.add_button("Jenkins build", args.jenkins_build_url)
            # TODO: Add a 'Rebuild in Jenkins' button also (which will be possible only if ditch our 'buildWithParameters' pipelines so we can use the '/build' endpoint )


class PullRequestUpdateCard(TeamsCardTemplate):
    def __init__(self, args):
        required_arguments = {'pull_request_id'}
        optional_arguments = {'pull_request_author'}
        super().__init__(required_arguments, optional_arguments, args)

    def build(self, args):
        self.set_title(f"Pull Request {args.pull_request_id} [update]")
        # This simple card only shows the 'PR requestor' fact.
        if hasattr_nonempty_str(args, 'pull_request_author'):
            self.add_section().add_fact("Requested by", args.pull_request_author)


class BuildFailCard(TeamsCardTemplate):
    def __init__(self, args): 
        required_arguments = {'jenkins_build_id', 'jenkins_build_url', 'status', 'build_type'}
        optional_arguments = {'jenkins_job_name', 'pipeline_name', 'playbook_limit', 'pull_request_id', 'pull_request_url'}
        super().__init__(required_arguments, optional_arguments, args)

    def build(self, args):
        self.set_title(f'{args.build_type} {args.status.capitalize()}')
        build_facts = self.add_section()
        build_facts.add_fact("Jenkins Build", f'[Build #{args.jenkins_build_id}]({args.jenkins_build_url})')
        if hasattr_nonempty_str(args, 'pull_request_id') and hasattr_nonempty_str(args, 'pull_request_url'):
            build_facts.add_fact("Pull Request", f"[PR #{args.pull_request_id}]({args.pull_request_url})")
        if hasattr_nonempty_str(args, 'playbook_limit'):  # Optional
            build_facts.add_fact("Playbook limit", args.playbook_limit)
        if hasattr_nonempty_str(args, 'jenkins_job_name'):  # Optional
            pipeline_url = args.jenkins_build_url.rsplit('/',2)[0]
            pipeline_name = args.jenkins_job_name.replace('/',' ︙ ')
            build_facts.add_fact("Jenkins Pipeline", f"[{pipeline_name}]({pipeline_url})")
        self.add_button(f'Review {args.status} {args.build_type}', args.jenkins_build_url)
        self.add_button('Review Console Text', args.jenkins_build_url + "/console")


class BuildSuccessCard(TeamsCardTemplate):
    def __init__(self, args):
        required_arguments = {'build_type', 'status'}
        optional_arguments = {'pull_request_id', 'pull_request_url', 'playbook_limit'}
        super().__init__(required_arguments, optional_arguments, args)

    def build(self, args):
        self.set_title(f'{args.build_type} {args.status.capitalize()}')
        build_facts = self.add_section()    
        if hasattr_nonempty_str(args, 'pull_request_id') and hasattr_nonempty_str(args, 'pull_request_url'):
            build_facts.add_fact("Pull Request", f"[PR #{args.pull_request_id}]({args.pull_request_url})")
        elif hasattr_nonempty_str(args, 'pull_request_id'):
            build_facts.add_fact("Pull Request", f"PR #{args.pull_request_id}")
        if hasattr_nonempty_str(args, 'playbook_limit'):  # playbook_limit is optional
            pretty_playbook_limit = args.playbook_limit.replace(',', ' ︙ ')
            build_facts.add_fact("Playbook limit", pretty_playbook_limit)


class TeamsCardFactory:
    """Generates preformatted TeamsCard objects. This class is just a container for a bunch of static
    methods that parse 'args' (generally provided from python argparse) and determine what type of 
    TeamsCardTemplate should be built.
    """
    # TODO Could have a 'def register_card(message_type,additional_condition=foo())' 
    @staticmethod
    def build(args, message_type: TeamsMessageTypes):
        """Create a preformatted card -- the type of card built is based args provided.

        Args:
            args (argparse.Namespace): An 'arguments' object that contains all of the details required to 
            build a particular TeamsCard.  
            message_type (TeamsMessageTypes): Determines which type of TeamsCard to return -- depending
            on the message type, additional arguments will be required as well.

        Raises:
            ValueError: If the message_type specified is invalid.
            TypeError: If there are missing arguments from `args`

        Returns:
            TeamsCardTemplate: A Teams Card, built based on the arguments provided.
        """
        card = TeamsCardFactory._build(args, message_type)
        TeamsCardFactory._add_messages(args, card)
        return card

    @staticmethod
    def _add_messages(args, card):
        """Add 'messages' that are found in `args` to `card`.

        :param args: The args to be scanned for `message` or `file`.
        :param card: The card that will be updated with any 'messages' found in `args`.
        :return: The card.
        """
        if hasattr_nonempty_str(args, 'message') and len(args.message) > 0:
            TeamsCardFactory._add_message(args, card, args.message)
        if hasattr_nonempty_str(args, 'file') and len(args.file) > 0:
            TeamsCardFactory._add_message(args, card, ntsbuildtools.io.readfile(args.file))
        return card

    @staticmethod
    def _add_message(args, card, message):
        """Add the message to the card. Formatting details may be specified in args.

        :param args: The args to be scanned for formatting details, e.g. `max_size`, and `code_markdown`
        :param card: The card that will be updated with the `message`.
        :param message: The message to be added to the `card`.
        :return: The card.
        """
        content = MarkdownComment()

        trimmed = False
        if args.max_size and len(message) > args.max_size:
            message = message[0:args.max_size]
            trimmed = True

        if hasattr(args, 'code_markdown') and args.code_markdown:
            content.add_code_markdown(message)
        else:
            content.add_message(message)

        if trimmed:
            content.add_message("\n> *Message trimmed. **See Pull Request** for full message...*")

        card.add_section(str(content))

    @staticmethod
    def _build(args, message_type):
        # Handle the 'build status' message types
        if message_type == TeamsMessageTypes.BUILD_STATUS.value: 
            return TeamsCardFactory._build_build_status_card(args)
        # Handle the 'pull request' message types
        if message_type == TeamsMessageTypes.PULL_REQUEST.value:
            return TeamsCardFactory._build_pull_request_card(args)
        # If the message type doesn't align with what we've got, should we just return an empty card or throw an error?
        raise ValueError(f'Expected message_type to be one of: {", ".join(TeamsMessageTypes.values())}')

    @staticmethod
    def _build_build_status_card(args):
        # Check that we have required args to at least execute this factory method.
        if not hasattr_nonempty_str(args, 'status'):
            raise TypeError(f'To create a Card with message-type={TeamsMessageTypes.BUILD_STATUS}, the `status` arg must be provided.')
        # Set the 'build_type' string based on the message_type
        if args.status.upper() == 'SUCCESS':
            return BuildSuccessCard(args)
        else:
            return BuildFailCard(args)

    @staticmethod
    def _build_pull_request_card(args):
        if hasattr(args, 'pull_request_update') and args.pull_request_update is True:
            return PullRequestUpdateCard(args)
        else: 
            return PullRequestCard(args)



"""Posts a preformatted card (aka 'actionable message', or 'MessageCard') to a Microsoft Teams channel.

This script supports specific preformatted messages oriented towards supporting Jenkins build processes."""
import configargparse
import requests
import logging
import sys

import ntsbuildtools.subcommands
import ntsbuildtools.teams.cards

logger = logging.getLogger(__name__)


def config_parser(parser):
    BUILD_STATUSES = ['STARTED', 'SUCCESS',
                      'UNSTABLE', 'FAILURE', 'NOT_BUILD', 'ABORTED']
    parent_parser = configargparse.ArgParser(add_help=False)
    parent_parser.add(
        '--webhook-url', '-u', 
        required=True, 
        env_var='TEAMS_WEBHOOK_URL',
        help='The Teams "Incoming Webhook" URL for a Teams channel (where messages will be posted).'
    )
    parent_parser.add(
        '--bitbucket-url',
        env_var='BITBUCKET_URL', 
        help="URL for Bitbucket."
    )
    parent_parser.add(
        '--message',
        help='Custom messgage to be printed at the bottom of the card.'
    )
    parent_parser.add(
        '--file', 
        help='''Custom message to be printed at the bottom of the card, loaded from the
                file-path indicated by this argument.'''
    )
    parent_parser.add(
        '--diff-markdown', 
        action='store_true',
        help='Wrap the custom message(s) in ```diff\n{message}\n``` markdown.'
    )
    parent_parser.add(
        '--code-markdown', 
        action='store_true',
        help='Wrap the custom message(s) in ```\n{message}\n``` markdown.'
    )
    parent_parser.add(
        '--max-size', 
        '--trim', 
        dest='max_size', 
        type=int, 
        default=24000,  # 28 KB is the limit today -- leave 4 KB for our meta-data + Teams meta-data.
        help='''Truncate the message, to avoid surpassing Teams "Channel conversation post size" limitation 
                (See https://docs.microsoft.com/en-us/microsoftteams/limits-specifications-teams#chat).'''
    )

    subparsers = parser.add_subparsers(
        title="message-type", dest='message_type',
        description="There are built in 'message types' that can be sent."
    )
    subparsers.required = True

    # Subparser for Build Status Cards
    build_status = subparsers.add_parser(
        ntsbuildtools.teams.cards.TeamsMessageTypes.BUILD_STATUS.value, 
        parents=[parent_parser], help='Generate a "Jenkins build status" message.'
    )
    build_status.add('status', choices=BUILD_STATUSES,
                     help='Provide the build status here (from the provided choices).')
    build_status.add('--build-type', default='BUILD', choices=['BUILD', 'DEPLOYMENT'],
                     help="The 'type' of the build (from provided choices).")
    build_status.add('--jenkins-build-url', env_var='BUILD_URL',
                     help='The URL that links directly to the Jenkins build.')
    build_status.add('--pull-request-url', env_var='PR_URL',
                     help='The URL that links directly to the Pull Request to which this build is associated.')
    build_status.add('--pull-request-id', env_var='PR_ID',
                     help='The ID of the Bitbucket pull request.')
    build_status.add('--jenkins-job-name', env_var='JOB_NAME',
                     help='The name of the Job in Jenkins.')
    build_status.add('--jenkins-build-id', env_var='BUILD_ID',
                     help='The ID of this particular build in Jenkins.')
    build_status.add('--playbook-limit', env_var='PLAYBOOK_LIMIT',
                     help='A string indicating the PLAYBOOK_LIMIT for this build.')

    # Subparser for Pull Request Cards
    pull_request = subparsers.add_parser(
        ntsbuildtools.teams.cards.TeamsMessageTypes.BUILD_STATUS.PULL_REQUEST.value, 
        parents=[parent_parser], help='Generate a "Bitbucket Pull Request" message.'
    )
    pull_request.add('--pull-request-id', env_var='PR_ID',
                     help='The ID of the Bitbucket pull request.')
    pull_request.add('--pull-request-update', action='store_true', env_var='PR_UPDATE',
                     # TODO Document the following: Assign "true" to PR_UPDATE env variable, per configargparse docs
                     help='Provide this flag if this is a pull request update (it will produce a much smaller/simpler Teams Card).')
    pull_request.add('--pull-request-author', env_var='PR_AUTHOR',
                     help='The person who submitted the Bitbucket pull request.')
    pull_request.add('--project', env_var='BITBUCKET_PROJECT',
                     help='The Bitbucket project key for the project where the pull request exists.')
    pull_request.add('--repo-slug', env_var='BITBUCKET_REPO',
                     help='The Bitbucket repository slug.')
    pull_request.add('--pull-request-title', env_var='PR_TITLE',
                     help='The title of the Bitbucket pull request.')
    pull_request.add('--pull-request-dest', env_var='PR_DESTINATION',
                     help='The destination branch of the Bitbucket pull request.')
    pull_request.add('--pull-request-description', env_var='PR_DESCRIPTION',
                     help='The description of the Bitbucket pull request.')
    pull_request.add('--playbook-limit', env_var='PLAYBOOK_LIMIT',
                     help='A string indicating the PLAYBOOK_LIMIT for this build.')
    pull_request.add('--build-status', choices=BUILD_STATUSES,
                     help='Provide the build status (from the provided choices).')
    pull_request.add('--jenkins-build-url', env_var='BUILD_URL',
                     help='The URL that links directly to the particular Jenkins build.')


def main(args):
    try:
        card = ntsbuildtools.teams.cards.TeamsCardFactory.build(args, args.message_type)
        response = requests.post(args.webhook_url, json=card.to_json())
        response.raise_for_status()
        _raise_for_http_413(response)
    except (requests.exceptions.HTTPError, ntsbuildtools.exceptions.BTTeamsCardFactoryError) as error:
        logger.exception(error)
        print(error)
        sys.exit(1)


def _raise_for_http_413(response):
    # Teams does not actually set the status_code for HTTP 413... so we parse the response.text to handle for it
    if 'HTTP error 413' in response.text:
        response.status_code = 413
        response.raise_for_status()



import json
import datetime
from typing import Collection
from dataclasses import dataclass

import ntsbuildtools.ansible.return_values.default
import ntsbuildtools.ansible.return_values.factory as return_values_factory
from ntsbuildtools.markdown.comments import MarkdownComment

@dataclass
class TaskResult:
    """Stores the results from an Ansible Task that executed against a specific Ansible host/target.
    """
    failed: bool
    host_name: str
    task_name: str
    ansible_module: str
    start_time: datetime.datetime
    end_time: datetime.datetime
    return_values: ntsbuildtools.ansible.return_values.default.ReturnValues

    def as_markdown(self):
        markdown = MarkdownComment()
        if self.failed:
            markdown.add_heading(3, f'[TASK] {self.task_name}')
            markdown.add_heading(4, f'{self.host_name} ***(ERROR)***')
            markdown.add_message(self.return_values.pretty_error())
        else:
            markdown.add_heading(3, f'[TASK] {self.task_name}')
            markdown.add_heading(4, f'{self.host_name} ***(diff)***')
            markdown.add_diff_markdown(self.return_values.pretty_diff())
        return str(markdown)


class PlaybookResults:
    def __init__(self):
        self._results = []
        self._results_by_task = {}

    @property
    def tasks(self):
        """Get the list of the task names that are associated with this PlaybookResults.

        Returns:
            List[str]: List of the tasks names associated to this PlaybookResults.
        """
        return self._results_by_task.keys()

    def add(self, task_result: TaskResult):
        """Add a TaskResult to this collection.

        TaskResults should be added in the same order as they were executed.

        Args:
            task_result (TaskResult): A TaskResult to be added to this collection.
        """
        # 1. Store in a flat data structure
        self._results.append(task_result)
        # 2. Store in task-nested data structure
        if task_result.task_name not in self._results_by_task:
            self._results_by_task[task_result.task_name] = []
        self._results_by_task[task_result.task_name].append(task_result)

    def has_failures(self):
        return any(task_result.failed for task_result in self._results)

    def failures(self) -> Collection[TaskResult]:
        return list(filter(lambda task_result: task_result.failed, self._results))

    def by_task_name(self, task_name) -> Collection[TaskResult]:
        return self._results_by_task[task_name]
        # return filter(lambda task_result: task_result.task_name == task_name, self._results)

    def by_ansible_module(self, ansible_module):
        return list(filter(lambda task_result: task_result.ansible_module == ansible_module, self._results))

    @classmethod
    def from_json(cls, json_payload):
        """Take the json output of an ansible-playbook run and process it into a PlaybookResults data structure.

        Args:
            json_payload: The raw json output from running ansible-playbook. 
            This can be configured in the ansible.cfg file under [Defaults] > stdout_callback = json
        Requirements:
            Playbook runs need to be ran with --diff to get the correct output.

        Raises:
            JSONDecodeError: If the json_payload is not well formed.
        """
        output = json.loads(json_payload)

        # We should only have one play here, so let's make that assumption and allow the index error to take place otherwise.
        play = output['plays'][0]
        return cls._parse_playbook_results(play)

    @classmethod
    def _parse_playbook_results(cls, play):
        results = cls()
        for task in play['tasks']:
            for host_name, return_values in task['hosts'].items():
                task_details = task['task']
                results.add(cls._parse_task_result(return_values, host_name, task_details))
        return results

    @staticmethod
    def _parse_task_result(return_values_raw, host_name, task_details):
        # Surprising: the 'Ansible Module' associated to the task is not provided in task_details... 
        #   Instead, it is found as the `action` property from the return_values.
        ansible_module = return_values_raw['action']

        # Build a more useful return_values object to operate on
        return_values = return_values_factory.build(return_values_raw, ansible_module)

        # Generate the TaskResult data structure
        return TaskResult(return_values=return_values, failed=return_values.failed, 
                          ansible_module=ansible_module, host_name=host_name, task_name=task_details['name'],
                          start_time=task_details['duration']['start'], end_time=task_details['duration']['end'])
